import ast
import operator
from astoptimizer import UNSET
from astoptimizer.ast_tools import ast_contains, copy_lineno, new_constant
from astoptimizer.config import optimize_unicode
from astoptimizer.compatibility import (
    u,
    PYTHON2, PYTHON3,
    is_bytes_ascii, is_singleton,
    INT_TYPES, FLOAT_TYPES, COMPLEX_TYPES,
    NATIVE_STR, BYTES_TYPE, UNICODE_TYPE, STR_TYPES,
    IMMUTABLE_ITERABLE_TYPES)
import sys

def operator_not_in(data, item):
    return item not in data

EVAL_COMPARE = {
    # a == b, a != b
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    # a < b, a <= b
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    # a > b, a >= b
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    # a in b, a not in b
    ast.In: operator.contains,
    ast.NotIn: operator_not_in,
    # a is b, a is not b
    ast.Is: operator.is_,
    ast.IsNot: operator.is_not,
}

EVAL_UNARYOP = {
    # not a, ~a, +a, -a
    ast.Not: operator.not_,
    ast.Invert: operator.invert,
    ast.UAdd: operator.pos,
    ast.USub: operator.neg,
}

EVAL_BINOP = {
    # a + b, a - b, a * b
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    # see binop(): floordiv() may be used for int/int on Python 2
    ast.Div: operator.truediv,
    ast.FloorDiv: operator.floordiv,
    ast.Mod: operator.mod,
    # a ** b
    ast.Pow: operator.pow,
    # a << b, a >> b
    ast.LShift: operator.lshift,
    ast.RShift: operator.rshift,
    # a & b, a | b, a ^ b
    ast.BitAnd: operator.and_,
    ast.BitOr: operator.or_,
    ast.BitXor: operator.xor,
}

FLOAT_BINOPS = (
    ast.Add, ast.Sub,
    ast.Mult, ast.Div, ast.FloorDiv, ast.Mod,
    ast.Pow)

DIVIDE_BINOPS = (ast.Div, ast.FloorDiv, ast.Mod)
COMPLEX_BINOPS = (
    ast.Add, ast.Sub,
    ast.Mult, ast.Div)
if PYTHON2:
    COMPLEX_BINOPS += (ast.FloorDiv, ast.Mod)

TUPLE_BINOPS = (ast.Add,)
FROZENSET_BINOPS = (ast.BitAnd, ast.BitOr, ast.BitXor, ast.Sub)

class Namespace:
    def __init__(self):
        self._aliases_enabled = True
        self.qualnames = {}
        self._removed_aliases = set()

        self._vars_enabled = False
        self.values = {}
        self._unassigned = set()

    @property
    def vars_enabled(self):
        return self._vars_enabled

    def enable_experimental_vars(self):
        self._vars_enabled = True

    def copy(self):
        ns = Namespace()
        ns._aliases_enabled = self._aliases_enabled
        ns.qualnames = self.qualnames.copy()
        ns._removed_aliases = self._removed_aliases

        # don't copy variables
        # ns._vars_enabled = self._vars_enabled
        # ns.values = self.values.copy()
        # ns._unassigned = self._unassigned.copy()
        return ns

    def add_alias(self, qualname, name):
        """
        Register name as an alias to qualname.

        If name is already registered, unregister the previous
        alias. If the name was already removed, don't register
        the new name.
        """
        if not self._aliases_enabled:
            return
        if name in self._removed_aliases:
            return
        if name in self.qualnames:
            if self.qualnames[name] == qualname:
                return
            self._removed_aliases.add(name)
            del self.qualnames[name]
            return
        self.qualnames[name] = qualname

    def disable_aliases(self):
        self._aliases_enabled = False
        self.qualnames.clear()

    def get_qualname(self, name):
        if not self._aliases_enabled:
            return UNSET
        if name in self.qualnames:
            return self.qualnames[name]
        fullname = name
        prefixes = []
        while '.' in name:
            prefix, name = name.split(".", 1)
            prefixes.append(prefix)
            key = '.'.join(prefixes)
            if key in self.qualnames:
                return self.qualnames[key] + '.' + name
        return fullname

    def assign(self, qualname, value):
        if not self._vars_enabled:
            return
        if qualname in self._unassigned:
            return
        self.values[qualname] = value

    def unassign(self, qualname):
        if not self._vars_enabled:
            return
        self._unassigned.add(qualname)
        self.values.pop(qualname, None)

    def disable_vars(self):
        self._vars_enabled = False
        self.values.clear()

    def get_var(self, name):
        if not self._vars_enabled:
            return UNSET
        qualname = self.get_qualname(name)
        if qualname is UNSET:
            return UNSET
        return self.values.get(qualname, UNSET)

def remove_duplicate_pass(node_list):
    # Remove duplicate pass instructions
    index = 1
    while index < len(node_list):
        if isinstance(node_list[index], ast.Pass):
            del node_list[index]
        else:
            index += 1
    if (len(node_list) > 1
    and isinstance(node_list[0], ast.Pass)):
        del node_list[0]

class Optimizer(ast.NodeTransformer):
    def __init__(self, config):
        ast.NodeTransformer.__init__(self)
        self.config = config
        self.namespace = None
        self.root = None
        self.is_conditional = None

    def _optimize(self, tree, namespace):
        self.namespace = namespace
        if self.config.use_experimental_vars:
            self.namespace.enable_experimental_vars()
        self.root = tree
        self.is_conditional = False
        return self.visit(tree)

    def optimize(self, tree):
        namespace = Namespace()
        return self._optimize(tree, namespace)

    def warning(self, message):
        self.config.warning_func(message)

    def get_constant(self, node):
        if node is None:
            return None
        if isinstance(node, ast.Num):
            return node.n
        if isinstance(node, ast.Str):
            text = node.s
            if (isinstance(text, UNICODE_TYPE)
            and not optimize_unicode(self.config, text)):
                return UNSET
            return text
        if PYTHON3:
            if isinstance(node, ast.Bytes):
                return node.s
        if isinstance(node, ast.Name):
            name = node.id
            qualname = self.namespace.get_qualname(name)
            if qualname is UNSET:
                return UNSET
            constant = self.config.get_constant(qualname)
            if constant is not UNSET:
                return constant
            return self.namespace.get_var(name)
        if isinstance(node, ast.Tuple):
            elts = node.elts
            if len(elts) > self.config.max_tuple_length:
                return UNSET
            constants = []
            for elt in elts:
                constant = self.get_constant(elt)
                if constant is UNSET:
                    return UNSET
                constants.append(constant)
            return tuple(constants)
        if (isinstance(node, ast.Call)
        and 'builtin_funcs' in self.config.features
        and isinstance(node.func, ast.Name)
        and node.func.id == 'frozenset'):
            if len(node.args) == 1:
                arg = self.get_constant(node.args[0])
                if arg is UNSET:
                    return UNSET
                if not isinstance(arg, IMMUTABLE_ITERABLE_TYPES):
                    return UNSET
                if len(arg) > self.config.max_tuple_length:
                    return UNSET
                return frozenset(arg)
            elif len(node.args) == 0:
                return frozenset()
        return UNSET

    def get_constant_list(self, nodes):
        constants = []
        for node in nodes:
            constant = self.get_constant(node)
            if constant is UNSET:
                return UNSET
            constants.append(constant)
        return constants

    def can_drop(self, node_list):
        return True

    def remove_dead_code(self, node_list):
        # Remove dead code
        # Example: "return 1; return 2" => "return 1"
        truncate = None
        for index, node in enumerate(node_list[:-1]):
            if not isinstance(node, (ast.Return, ast.Raise)):
                continue
            if not self.can_drop(node_list[index+1:]):
                continue
            truncate = index
            break
        if truncate is None:
            return
        if truncate == len(node_list) - 1:
            return
        del node_list[truncate+1:]

    def optimize_node_list(self, node_list):
        if not self.config.remove_dead_code:
            return
        remove_duplicate_pass(node_list)
        self.remove_dead_code(node_list)

    def generic_visit(self, node):
        for field, old_value in ast.iter_fields(node):
            old_value = getattr(node, field, None)
            if isinstance(old_value, list):
                new_values = self.visit_list(old_value)
                old_value[:] = new_values
            elif isinstance(old_value, ast.AST):
                new_node = self.visit(old_value)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        return node

    def visit(self, node, conditional=None):
        old_is_conditional = self.is_conditional
        try:
            if conditional:
                self.is_conditional = True
            return self._visit(node)
        finally:
            self.is_conditional = old_is_conditional

    def _visit(self, node):
        method = 'fullvisit_' + node.__class__.__name__
        visitor = getattr(self, method, None)
        if visitor is not None:
            new_node = visitor(node)
            if new_node is not None:
                return new_node
            else:
                return node

        new_node = self.generic_visit(node)

        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, UNSET)
        if visitor is UNSET:
            return new_node

        optimized_node = visitor(new_node)
        if optimized_node is not None:
            return optimized_node
        else:
            return new_node

    def visit_list(self, node_list, conditional=False):
        new_node_list = []
        for node in node_list:
            if not isinstance(node, ast.AST):
                new_node_list.append(node)
                continue

            new_node = self.visit(node, conditional=conditional)
            if new_node is None:
                continue
            elif isinstance(new_node, ast.AST):
                new_node_list.append(new_node)
            elif isinstance(new_node, list):
                assert all(isinstance(node, ast.AST) for node in new_node)
                new_node_list.extend(new_node)
            else:
                raise TypeError(type(new_node))
        self.optimize_node_list(new_node_list)
        return new_node_list

    def visit_Name(self, node):
        constant = self.get_constant(node)
        if constant is UNSET:
            return
        return new_constant(node, constant)

    def get_attribute_name(self, node):
        if isinstance(node.value, ast.Name):
            name = node.value.id
        elif isinstance(node.value, ast.Attribute):
            name = self.get_attribute_name(node.value)
            if name is UNSET:
                return UNSET
        else:
            return UNSET
        return "%s.%s" % (name, node.attr)

    def visit_Attribute(self, node):
        name = self.get_attribute_name(node)
        if name is UNSET:
            return
        name = self.namespace.get_qualname(name)
        if name is UNSET:
            return
        constant = self.config.get_constant(name)
        if constant is UNSET:
            return
        return new_constant(node, constant)

    def check_format_args(self, format, values):
        if PYTHON2:
            if isinstance(format, BYTES_TYPE):
                if is_bytes_ascii(format):
                    return True
                # b'x\xe9=%s' % u'abc' gives a different result
                # depending on the default encoding
                return all(not isinstance(value, UNICODE_TYPE) for value in values)
            elif isinstance(format, UNICODE_TYPE):
                return all(
                    (not isinstance(value, BYTES_TYPE)) or is_bytes_ascii(value)
                    for value in values)
        else:
            # Python 3
            return all(not isinstance(value, BYTES_TYPE) for value in values)
        return True

    def string_format(self, node, format, arg):
        if isinstance(arg, tuple):
            values = arg
        else:
            values = (arg,)
        if not self.check_format_args(format, values):
            return UNSET
        try:
            result = format % arg
        except (TypeError, ValueError, UnicodeError):
            # format string or arguments are invalid
            return UNSET
        if not self.config.check_result(result):
            return UNSET
        return new_constant(node, result)

    def check_binop_cst(self, op, left, right):
        if (isinstance(left, COMPLEX_TYPES)
        and isinstance(right, COMPLEX_TYPES)):
            if (isinstance(op, DIVIDE_BINOPS)
            and not right):
                # division by zero
                return False

            if (isinstance(op, ast.Pow)
            and isinstance(left, FLOAT_TYPES)
            and isinstance(right, FLOAT_TYPES)):
                # -1.0 ** 0.5 raises a ValueError
                if right < 1.0 and right != 0.0 and left < 0:
                    return False

            if (isinstance(op, (ast.LShift, ast.RShift))
            and right < 0):
                # 1 << -3 and 1 >> -3 raise a ValueError
                return False

        if (isinstance(left, INT_TYPES)
        and isinstance(right, INT_TYPES)):
            return True

        if (isinstance(left, FLOAT_TYPES)
        and isinstance(right, FLOAT_TYPES)):
            return isinstance(op, FLOAT_BINOPS)

        if (isinstance(left, COMPLEX_TYPES)
        and isinstance(right, COMPLEX_TYPES)):
            return isinstance(op, COMPLEX_BINOPS)

        if isinstance(op, ast.Mult):
            if isinstance(right, INT_TYPES):
                # str * int
                if isinstance(left, STR_TYPES):
                    return (len(left) * right <= self.config.max_string_length)
                # tuple * int
                if isinstance(left, tuple):
                    return (len(left) * right <= self.config.max_tuple_length)

            if isinstance(left, INT_TYPES):
                # int * str
                if isinstance(right, STR_TYPES):
                    return (left * len(right) <= self.config.max_string_length)
                # int * tuple
                if isinstance(right, tuple):
                    return (left * len(right) <= self.config.max_tuple_length)

        if (isinstance(left, tuple)
        and isinstance(right, tuple)):
            return isinstance(op, TUPLE_BINOPS)

        if (isinstance(left, frozenset)
        and isinstance(right, frozenset)):
            return isinstance(op, FROZENSET_BINOPS)

        return False

    def binop(self, node, eval_binop, left, left_cst, right, right_cst):
        if left_cst is UNSET:
            return UNSET
        if right_cst is UNSET:
            return UNSET

        if (isinstance(node.op, ast.Mod)
        and isinstance(left_cst, STR_TYPES)):
            # str % args
            if PYTHON3 and isinstance(left_cst, BYTES_TYPE):
                return UNSET
            return self.string_format(node, left_cst, right_cst)

        if (PYTHON2
        and isinstance(node.op, ast.Div)
        and (isinstance(left_cst, INT_TYPES)
             or isinstance(right_cst, INT_TYPES))):
            # Don't optimize a / b on Python 2, because it depends on the
            # -Q command line option, except if pythonenv feature is enable
            if 'pythonenv' not in self.config.features:
                return UNSET
            if not sys.flags.division_new:
                eval_binop = operator.floordiv

        ok = self.check_binop_cst(node.op, left_cst, right_cst)
        if not ok:
            return UNSET

        result = eval_binop(left_cst, right_cst)
        if not self.config.check_result(result):
            return UNSET
        return new_constant(node, result)

    def visit_BinOp(self, node):
        eval_binop = EVAL_BINOP.get(node.op.__class__)
        if not eval_binop:
            return
        left = node.left
        left_cst = self.get_constant(left)
        right = node.right
        right_cst = self.get_constant(right)
        new_node = self.binop(node, eval_binop, left, left_cst, right, right_cst)
        if new_node is not UNSET:
            return new_node
        return node

    def visit_BoolOp(self, node):
        if (not hasattr(node, 'op')
        and 'cpython_tests' in self.config.features):
            # test_compile of Python 2.7 tries to compile an invalid AST tree:
            # ast.BoolOp() node without any attribute
            return
        if not isinstance(node.op, (ast.And, ast.Or)):
            return
        if len(node.values) < 2:
            return
        values = node.values
        constants = []
        for value in values:
            constant = self.get_constant(value)
            if constant is UNSET:
                break
            constants.append(constant)
        else:
            # all arguments are constant
            result = constants[0]
            if isinstance(node.op, ast.And):
                for constant in constants[1:]:
                    result = result and constant
            else:
                for constant in constants[1:]:
                    result = result or constant
            return new_constant(node, result)

        result = values
        for index, constant in enumerate(constants):
            test = constant
            if isinstance(node.op, ast.Or):
                test = not constant
            if test:
                result = values[index+1:]
            else:
                return new_constant(node, constant)
        if len(result) > 1:
            new_node = ast.BoolOp(node.op, result)
            return copy_lineno(node, new_node)
        else:
            return result[0]

    def check_unary_op(self, op, value):
        if isinstance(op, ast.Not):
            return True
        if isinstance(op, ast.Invert):
            return isinstance(value, INT_TYPES)
        if isinstance(op, (ast.UAdd, ast.USub, ast.Invert)):
            return isinstance(value, COMPLEX_TYPES)
        return True

    def compare_generic(self, node, operand):
        if len(operand.ops) != 1:
            return UNSET
        op = operand.ops[0]
        if isinstance(op, ast.In):
            # Replace "not(a in b)" with "a not in b"
            operand.ops[0] = ast.NotIn()
            return operand
        elif isinstance(op, ast.Is):
            # Replace "not(a is b)" with "a is not b"
            operand.ops[0] = ast.IsNot()
            return operand
        elif isinstance(op, ast.NotIn):
            # Replace "not(a not is b)" with "a is b"
            operand.ops[0] = ast.In()
            return operand
        elif isinstance(op, ast.IsNot):
            # Replace "not(a not in b)" with "a in b"
            operand.ops[0] = ast.Is()
            return operand
        else:
            return UNSET

    def visit_UnaryOp(self, node):
        eval_unaryop = EVAL_UNARYOP.get(node.op.__class__)
        if eval_unaryop is None:
            return
        operand = node.operand
        constant = self.get_constant(operand)
        if constant is not UNSET:
            if not self.check_unary_op(node.op, constant):
                return
            result = eval_unaryop(constant)
            if not self.config.check_result(result):
                return
            return new_constant(node, result)
        elif (isinstance(node.op, ast.Not)
        and isinstance(operand, ast.Compare)):
            new_node = self.compare_generic(node, operand)
            if new_node is not UNSET:
                return new_node

    def subscript_slice(self, node, value, lower, upper, step):
        lower_cst = self.get_constant(lower)
        if lower_cst is UNSET:
            return UNSET
        upper_cst = self.get_constant(upper)
        if upper_cst is UNSET:
            return UNSET
        step_cst = self.get_constant(step)
        if step_cst is UNSET:
            return UNSET
        myslice = slice(lower_cst, upper_cst, step_cst)
        return new_constant(node, value[myslice])

    def subscript_index(self, node, value, index):
        index_constant = self.get_constant(index)
        if index_constant is UNSET:
            return UNSET
        if not isinstance(index_constant, INT_TYPES):
            return UNSET
        if index_constant >= 0:
            if index_constant >= len(value):
                return UNSET
        else:
            if (len(value) + index_constant) < 0:
                return UNSET
        result = value[index_constant]
        return new_constant(node, result)

    def visit_Subscript(self, node):
        value = node.value
        value_constant = self.get_constant(value)
        if value_constant is UNSET:
            return
        if isinstance(node.slice, ast.Slice):
            lower = node.slice.lower
            upper = node.slice.upper
            step = node.slice.step
            new_node = self.subscript_slice(node, value_constant, lower, upper, step)
            if new_node is not UNSET:
                return new_node
        elif isinstance(node.slice, ast.Index):
            index = node.slice.value
            new_node = self.subscript_index(node, value_constant, index)
            if new_node is not UNSET:
                return new_node

    def check_compare_in(self, item, data):
        if isinstance(data, (tuple, frozenset)):
            return True
        if (isinstance(data, STR_TYPES)
        and type(item) == type(data)):
            return True
        return False

    def check_compare(self, left, op, right):
        if isinstance(op, (ast.Is, ast.IsNot)):
            return (is_singleton(self.config, left)
                    and is_singleton(self.config, right))

        if (isinstance(left, COMPLEX_TYPES)
        and isinstance(right, COMPLEX_TYPES)):
            return True

        if (isinstance(left, STR_TYPES + (tuple, frozenset))
        and type(right) == type(left)):
            return True

        return False

    def compare(self, node, op, left, right):
        eval_compare = EVAL_COMPARE.get(op.__class__)
        if eval_compare is None:
            return UNSET
        if isinstance(op, (ast.In, ast.NotIn)):
            if not self.check_compare_in(left, right):
                return UNSET
            result = eval_compare(right, left)
        else:
            if not self.check_compare(left, op, right):
                return UNSET
            result = eval_compare(left, right)
        if not self.config.check_result(result):
            return UNSET
        return new_constant(node, result)

    def visit_Compare(self, node):
        left = node.left
        left_cst = self.get_constant(left)
        if left_cst is UNSET:
            return
        # FIXME: implement 1 < 2 < 3
        if len(node.ops) != 1:
            return
        op = node.ops[0]

        if len(node.comparators) != 1:
            return
        right = node.comparators[0]
        right_cst = self.get_constant(right)
        if right_cst is UNSET:
            return

        new_node = self.compare(node, op, left_cst, right_cst)
        if new_node is UNSET:
            return
        return new_node

    def if_block(self, parent, node_list):
        # Create a pass instruction if needed.
        # Example: if 0: print("debug") => pass
        if len(node_list) == 0:
            new_node = ast.Pass()
            return copy_lineno(parent, new_node)
        else:
            return node_list

    def fullvisit_If(self, node):
        if (not hasattr(node, 'test')
        and 'cpython_tests' in self.config.features):
            # test_compile of Python 2.7 tries to compile an invalid AST tree:
            # ast.If() node without any attribute
            return
        node.test = self.visit(node.test)

        drop = False
        if self.config.remove_dead_code:
            constant = self.get_constant(node.test)
            if constant is not UNSET:
                if constant:
                    check_node = node.orelse
                else:
                    check_node = node.body
                drop = self.can_drop(check_node)

        if drop:
            if constant:
                node.body = self.visit_list(node.body, conditional=True)
                return self.if_block(node, node.body)
            else:
                node.orelse = self.visit_list(node.orelse, conditional=True)
                return self.if_block(node, node.orelse)
        else:
            node.body = self.visit_list(node.body, conditional=True)
            node.orelse = self.visit_list(node.orelse, conditional=True)

    def visit_IfExp(self, node):
        constant = self.get_constant(node.test)
        if constant is UNSET:
            return
        if constant:
            return node.body
        else:
            return node.orelse

    def fullvisit_While(self, node):
        # Call can_drop() before visit_list(node.body) because
        # FunctionOptimizer sets its seen_yield attribute
        can_drop = self.can_drop(node.body)

        # visit body and orelse before test, because test may use a variable
        # unassigned during visit of body or orelse
        node.body = self.visit_list(node.body, conditional=True)
        node.orelse = self.visit_list(node.orelse, conditional=True)
        node.test = self.visit(node.test)

        if not self.config.remove_dead_code:
            return

        constant = self.get_constant(node.test)
        if constant is UNSET:
            return
        if constant:
            return
        # if not self.can_drop(node.body):
        if not can_drop:
            return

        # while False: ...
        return self.if_block(node, node.orelse)

    def print_func(self, node, args):
        constants = self.get_constant_list(args)
        if constants is UNSET:
            return UNSET
        if any(isinstance(constant, BYTES_TYPE)
               for constant in constants):
            return UNSET
        text = ' '.join(str(constant) for constant in constants)
        text = new_constant(node, text)
        func = ast.Name(id='print', ctx=ast.Load())
        new_node = ast.Call(func, [text], [], None, None)
        return copy_lineno(node, new_node)

    def call_func(self, node, name, args):
        if PYTHON3 and name == 'print':
            return self.print_func(node, args)
        if name not in self.config.functions:
            return UNSET
        func = self.config.functions[name]
        constants = self.get_constant_list(args)
        if constants is UNSET:
            return UNSET
        result = func.call(self.config, constants)
        if result is UNSET:
            return UNSET
        return new_constant(node, result)

    def call_method(self, node, func, args):
        obj_cst = self.get_constant(func.value)
        if obj_cst is UNSET:
            return UNSET
        attr = func.attr
        assert isinstance(attr, NATIVE_STR)

        obj_type = type(obj_cst)
        if obj_type not in self.config.methods:
            return UNSET
        methods = self.config.methods[obj_type]
        if attr not in methods:
            return UNSET
        method = methods[attr]

        constants = self.get_constant_list(args)
        if constants is UNSET:
            return UNSET
        result = method.call(self.config, obj_cst, constants)
        if result is UNSET:
            return UNSET
        return new_constant(node, result)

    def visit_Call(self, node):
        func = node.func
        args = node.args
        keywords = node.keywords
        starargs = node.starargs
        kwargs = node.kwargs
        # Don't support keywords, *args, **kw yet
        if keywords or starargs or kwargs:
            return
        if isinstance(func, ast.Name):
            name = self.namespace.get_qualname(func.id)
            if name is UNSET:
                return
            new_node = self.call_func(node, name, args)
            if new_node is not UNSET:
                return new_node
        elif isinstance(func, ast.Attribute) and isinstance(func.value, ast.Name):
            name = self.namespace.get_qualname(func.value.id)
            if name is UNSET:
                return
            name = "%s.%s" % (name, func.attr)
            new_node = self.call_func(node, name, args)
            if new_node is not UNSET:
                return new_node
        elif isinstance(func, ast.Attribute):
            new_node = self.call_method(node, func, args)
            if new_node is not UNSET:
                return new_node

    if PYTHON2:
        def visit_Print(self, node):
            if not node.values:
                return
            constants = self.get_constant_list(node.values)
            if constants is UNSET:
                return
            if any(isinstance(constant, UNICODE_TYPE)
                   for constant in constants):
                chunks = []
                for constant in constants:
                    if isinstance(constant, BYTES_TYPE):
                        if not is_bytes_ascii(constant):
                            return
                    else:
                        text = unicode(constant)
                    chunks.append(text)
                text = u(' ').join(chunks)
            else:
                text = ' '.join(str(constant) for constant in constants)
            text = new_constant(node, text)
            new_node = ast.Print(node.dest, [text], node.nl)
            return copy_lineno(node, new_node)

    def import_all(self, module_name):
        module = __import__(module_name, globals(), locals())
        for part in module_name.split(".")[1:]:
            module = getattr(module, part)
        if hasattr(module, '__all__'):
            all_attributes = module.__all__
        else:
            all_attributes = dir(module)
        all_attributes = set(all_attributes) - set(('__doc__', '__name__', '__package__'))
        for attr in all_attributes:
            qualname = "%s.%s" % (module_name, attr)
            self.namespace.add_alias(qualname, attr)

    def visit_ImportFrom(self, node):
        # from sys import platform: sys -> sys.platform
        # from sys import platform as plat: plat -> sys.platform
        module = node.module
        names = node.names
        for alias in names:
            assert isinstance(alias, ast.alias)
            if (not hasattr(alias, 'asname')
            and 'cpython_tests' in self.config.features):
                continue
            qualname = "%s.%s" % (module, alias.name)
            if alias.asname:
                name = alias.asname
            else:
                name = alias.name
            if name == '*':
                if self.config.is_module_used(module):
                    self.import_all(module)
                else:
                    self.warning('Disable optimizations because of "from %s import *"'
                                 % module)
                    self.namespace.disable_aliases()
                    return
            else:
                self.namespace.add_alias(qualname, name)

    def visit_Import(self, node):
        # import math as M: M -> math
        for alias in node.names:
            assert isinstance(alias, ast.alias)
            if not alias.asname:
                continue
            self.namespace.add_alias(alias.name, alias.asname)
        return node

    def assign_multiple(self, names, values):
        if len(names) != len(values):
            return False
        values = self.get_constant_list(values)
        if values is UNSET:
            return False
        for name, value in zip(names, values):
            self.namespace.assign(name, value)
        return True

    def _get_assign_name(self, main_node, node):
        if isinstance(node, ast.Name):
            # var = value
            return (node.id, True)
        elif isinstance(node, ast.Attribute):
            # var.attr = value, var1.attr1.attr2 = value
            result = self._get_assign_name(main_node, node.value)
            if result is None:
                return None
            name, supported = result
            if not supported:
                return None
            name = "%s.%s" % (name, node.attr)
            return (name, supported)
        elif isinstance(node, ast.Subscript):
            # var[index] = value, var[a:b] = value
            result = self._get_assign_name(main_node, node.value)
            if result is None:
                return None
            name, supported = result
            if not supported:
                return None
            return (name, False)
        elif isinstance(node, ast.Call):
            # func().attr = value
            return (None, False)

        # Not supported: disable optimizations on variables
        return None

    def assign_single(self, name, supported, value):
        if name is None:
            return

        if not supported:
            self.namespace.unassign(name)
            return

        value = self.get_constant(value)
        if value is UNSET:
            self.namespace.unassign(name)
            return
        self.namespace.assign(name, value)

    def assign(self, node, target, value, assign_supported):
        result = self._get_assign_name(node, target)
        if result is not None:
            # x = value
            name, supported = result
            self.assign_single(name, assign_supported and supported, value)
            return

        if (not isinstance(target, ast.Tuple)
        or len(target.elts) <= 1):
            self.disable_vars(node)
            return

        names = []
        supported = assign_supported
        for elt in target.elts:
            result = self._get_assign_name(node, elt)
            if result is None:
                self.disable_vars(node)
                return
            name, elt_supported = result
            names.append(name)
            supported &= elt_supported

        if supported and isinstance(value, ast.Tuple):
            # "x, y = value" or "x, y = 1, 2"
            values = value.elts
            if not self.assign_multiple(names, values):
                for name in names:
                    self.namespace.unassign(name)
        else:
            # Example: a, b = func()
            for name in names:
                if name is None:
                    continue
                self.namespace.unassign(name)

    if sys.version_info >= (3, 3):
        # Python 3.3+
        def fullvisit_With(self, node):
            # with
            for item in node.items:
                # don't visit optional_vars to not replace a variable name by its
                # value

                var = item.optional_vars
                if var is not None:
                    self.unassign(var)

            for item in node.items:
                item.context_expr = self.visit(item.context_expr)
            node.body = self.visit_list(node.body)

        def fullvisit_Try(self, node):
            # try/except/else/finally
            node.body = self.visit_list(node.body, True)
            node.handlers = self.visit_list(node.handlers, True)
            node.orelse = self.visit_list(node.orelse, True)
            node.finalbody = self.visit_list(node.finalbody, True)
    else:
        # Python 3.2 and older
        def fullvisit_With(self, node):
            # with
            var = node.optional_vars
            if var is not None:
                self.unassign(var)

            node.context_expr = self.visit(node.context_expr)
            # don't visit optional_vars to not replace a variable name by its
            # value
            node.body = self.visit_list(node.body)

        def fullvisit_TryFinally(self, node):
            # try/finally
            node.body = self.visit_list(node.body, True)
            node.finalbody = self.visit_list(node.finalbody, True)

        def fullvisit_TryExcept(self, node):
            # try/except/else
            node.body = self.visit_list(node.body, True)
            node.handlers = self.visit_list(node.handlers, True)
            node.orelse = self.visit_list(node.orelse, True)

    def fullvisit_Assign(self, node):
        if not self.config.use_experimental_vars:
            node.value = self.visit(node.value)
            return

        # FIXME: use two steps in the optimizer instead of using
        # such ugly hack
        supported = not self.is_conditional
        if supported:
            node.value = self.visit(node.value)
            for target in node.targets:
                self.assign(node, target, node.value, supported)
        else:
            for target in node.targets:
                self.assign(node, target, None, False)
            # Visit after unassign to avoid issues like:
            # "i = 0; # for x in range(n): i = i + 1"
            # Don't replace "i = i + 1" with "i = 0 + 1"
            node.value = self.visit(node.value)

    def fullvisit_AugAssign(self, node):
        node.value = self.visit(node.value)
        if self.config.use_experimental_vars:
            self.assign(node, node.target, None, False)

    def fullvisit_Delete(self, node):
        if not self.config.use_experimental_vars:
            return

        for target in node.targets:
            self.assign(node, target, None, False)

    def disable_vars(self, node):
        if self.namespace.vars_enabled:
            self.warning('Disable optimizations on variables: %s is not supported'
                         % ast.dump(node))
        self.namespace.disable_vars()

    def unassign(self, node):
        if isinstance(node, ast.Name):
            self.namespace.unassign(node.id)
        elif isinstance(node, ast.Tuple):
            for elt in node.elts:
                self.unassign(elt)
        else:
            self.disable_vars(node)

    def fullvisit_comprehension(self, node):
        self.unassign(node.target)
        node.iter = self.visit(node.iter)

    def fullvisit_For(self, node):
        #node.target = self.visit(node.target)
        self.unassign(node.target)

        node.iter = self.visit(node.iter)
        node.body = self.visit_list(node.body, conditional=True)
        node.orelse = self.visit_list(node.orelse, conditional=True)

        if (isinstance(node.iter, ast.Call)
        and isinstance(node.iter.func, ast.Name)):
            func_name = node.iter.func
            # for x in range(...) => for x in xrange(...)
            if (PYTHON2
            and func_name.id == 'range'):
                func_name.id = 'xrange'
        elif isinstance(node.iter, ast.List):
            iter_list = node.iter
            if len(iter_list.elts) > self.config.max_tuple_length:
                return
            node.iter = ast.Tuple(elts=iter_list.elts, ctx=iter_list.ctx)
            copy_lineno(iter_list, node.iter)

    def fullvisit_arguments(self, node):
        # Don't visit arguments
        if PYTHON3:
            for arg in node.args:
                self.namespace.unassign(arg.arg)
        else:
            for arg in node.args:
                self.unassign(arg)

    def fullvisit_FunctionDef(self, node):
        optimizer = FunctionOptimizer(self.config)
        func_namespace = self.namespace.copy()
        return optimizer._optimize(node, func_namespace)

    def fullvisit_ClassDef(self, node):
        if node is self.root:
            return self.generic_visit(node)
        else:
            optimizer = Optimizer(self.config)
            class_namespace = self.namespace.copy()
            return optimizer._optimize(node, class_namespace)


class FunctionOptimizer(Optimizer):
    def __init__(self, config):
        Optimizer.__init__(self, config)
        self.is_generator = None
        self.seen_yield = None

    def fullvisit_FunctionDef(self, node):
        if node is self.root:
            return self.generic_visit(node)
        else:
            return Optimizer.fullvisit_FunctionDef(self, node)

    def can_drop(self, node_list):
        if not self.is_generator:
            return True
        # Without "yield", a function is no more a generator.
        # "return 3" is a SyntaxError in a generator.
        # http://tomlee.co/2008/04/the-internals-of-python-generator-functions-in-the-ast/
        if self.seen_yield:
            # FIXME: drop "return" but not "return 3"
            return not ast_contains(node_list, ast.Return)
        else:
            # FIXME: handle seen_yield=None case for remove_dead_code()
            return not ast_contains(node_list, (ast.Yield, ast.Return))

    def remove_dead_code(self, node_list):
        seen = self.seen_yield
        try:
            self.seen_yield = None
            Optimizer.remove_dead_code(self, node_list)
        finally:
            self.seen_yield = seen

    def visit_Yield(self, node):
        self.seen_yield = True

    def _optimize(self, tree, namespace):
        self.is_generator = ast_contains(tree, ast.Yield)
        self.seen_yield = False
        return Optimizer._optimize(self, tree, namespace)

